@@ -1,51 +1,61 @@
+#ifndef ITHREAD_H
+#define ITHREAD_H
+
 /*******************************************************************************
  *
- * Copyright (c) 2000-2003 Intel Corporation 
- * All rights reserved. 
+ * Copyright (c) 2000-2003 Intel Corporation
+ * All rights reserved.
+ * Copyright (c) 2012 France Telecom All rights reserved.
  *
- * Redistribution and use in source and binary forms, with or without 
- * modification, are permitted provided that the following conditions are met: 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
  *
- * * Redistributions of source code must retain the above copyright notice, 
- * this list of conditions and the following disclaimer. 
- * * Redistributions in binary form must reproduce the above copyright notice, 
- * this list of conditions and the following disclaimer in the documentation 
- * and/or other materials provided with the distribution. 
- * * Neither name of Intel Corporation nor the names of its contributors 
- * may be used to endorse or promote products derived from this software 
+ * * Redistributions of source code must retain the above copyright notice,
+ * this list of conditions and the following disclaimer.
+ * * Redistributions in binary form must reproduce the above copyright notice,
+ * this list of conditions and the following disclaimer in the documentation
+ * and/or other materials provided with the distribution.
+ * * Neither name of Intel Corporation nor the names of its contributors
+ * may be used to endorse or promote products derived from this software
  * without specific prior written permission.
- * 
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
- * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR 
- * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
- * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
- * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
- * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR
+ * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
  * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  ******************************************************************************/
 
-#ifndef ITHREADH
-#define ITHREADH
+/*!
+ * \file
+ */
+
+#if !defined(_WIN32)
+	#include <sys/param.h>
+#endif
+
+#include "UpnpGlobal.h" /* For UPNP_INLINE, EXPORT_SPEC */
+#include "UpnpUniStd.h" /* for close() */
+
 #ifdef __cplusplus
 extern "C" {
 #endif
 
-
 #include <pthread.h>
-#ifndef WIN32
-	#include <unistd.h>
-#endif
 
-#if defined(__FreeBSD__) || defined(__OpenBSD__)
+#if defined(BSD) && !defined(__GNU__)
 	#define PTHREAD_MUTEX_RECURSIVE_NP PTHREAD_MUTEX_RECURSIVE
 #endif
 
-#ifdef PTHREAD_MUTEX_RECURSIVE
+
+#if defined(PTHREAD_MUTEX_RECURSIVE) || defined(__DragonFly__)
 	/* This system has SuS2-compliant mutex attributes.
 	 * E.g. on Cygwin, where we don't have the old nonportable (NP) symbols
 	 */
@@ -65,7 +75,11 @@ extern "C" {
 
 #define ITHREAD_CANCELED PTHREAD_CANCELED
 
-  
+
+#define ITHREAD_STACK_MIN PTHREAD_STACK_MIN
+#define ITHREAD_CREATE_DETACHED PTHREAD_CREATE_DETACHED
+#define ITHREAD_CREATE_JOINABLE PTHREAD_CREATE_JOINABLE
+
 /***************************************************************************
  * Name: ithread_t
  *
@@ -74,8 +88,9 @@ extern "C" {
  *      typedef to pthread_t.
  *      Internal Use Only.
  ***************************************************************************/
-typedef pthread_t ithread_t; 
-  
+typedef pthread_t ithread_t;
+
+
 /****************************************************************************
  * Name: ithread_attr_t
  *
@@ -84,19 +99,19 @@ typedef pthread_t ithread_t;
  *      typedef to pthread_attr_t
  *      Internal Use Only
  ***************************************************************************/
-typedef pthread_attr_t ithread_attr_t;	
+typedef pthread_attr_t ithread_attr_t;
 
 
 /****************************************************************************
  * Name: start_routine
  *
  *  Description:
- *      Thread start routine 
+ *      Thread start routine
  *      Internal Use Only.
  ***************************************************************************/
-typedef void * (*start_routine) (void *arg);
+typedef void (*start_routine)(void *arg);
+
 
-  
 /****************************************************************************
  * Name: ithread_cond_t
  *
@@ -116,7 +131,7 @@ typedef pthread_cond_t ithread_cond_t;
  *      typedef to pthread_mutexattr_t
  *      Internal Use Only
  ***************************************************************************/
-typedef pthread_mutexattr_t ithread_mutexattr_t;	
+typedef pthread_mutexattr_t ithread_mutexattr_t;
 
 
 /****************************************************************************
@@ -139,7 +154,7 @@ typedef pthread_mutex_t ithread_mutex_t;
  *      NOT USED
  *      Internal Use Only
  ***************************************************************************/
-typedef pthread_condattr_t ithread_condattr_t;	
+typedef pthread_condattr_t ithread_condattr_t;
 
 
 /****************************************************************************
@@ -150,7 +165,9 @@ typedef pthread_condattr_t ithread_condattr_t;
  *      typedef to pthread_rwlockattr_t
  *      Internal Use Only
  ***************************************************************************/
-typedef pthread_rwlockattr_t ithread_rwlockattr_t;	
+#if UPNP_USE_RWLOCK
+typedef pthread_rwlockattr_t ithread_rwlockattr_t;
+#endif /* UPNP_USE_RWLOCK */
 
 
 /****************************************************************************
@@ -161,7 +178,93 @@ typedef pthread_rwlockattr_t ithread_rwlockattr_t;
  *      typedef to pthread_rwlock_t
  *      Internal Use Only
  ***************************************************************************/
-typedef pthread_rwlock_t ithread_rwlock_t;	
+#if UPNP_USE_RWLOCK
+	typedef pthread_rwlock_t ithread_rwlock_t;
+#else
+	/* Read-write locks aren't available: use mutex instead. */
+	typedef ithread_mutex_t ithread_rwlock_t;
+#endif /* UPNP_USE_RWLOCK */
+
+
+/****************************************************************************
+ * Function: ithread_initialize_library
+ *
+ *  Description:
+ *      Initializes the library. Does nothing in all implementations, except
+ *      when statically linked for WIN32.
+ *  Parameters:
+ *      none.
+ *  Returns:
+ *      0 on success, Nonzero on failure.
+ ***************************************************************************/
+static UPNP_INLINE int ithread_initialize_library(void) {
+	int ret = 0;
+
+	return ret;
+}
+
+
+/****************************************************************************
+ * Function: ithread_cleanup_library
+ *
+ *  Description:
+ *      Clean up library resources. Does nothing in all implementations, except
+ *      when statically linked for WIN32.
+ *  Parameters:
+ *      none.
+ *  Returns:
+ *      0 on success, Nonzero on failure.
+ ***************************************************************************/
+static UPNP_INLINE int ithread_cleanup_library(void) {
+	int ret = 0;
+
+	return ret;
+}
+
+
+/****************************************************************************
+ * Function: ithread_initialize_thread
+ *
+ *  Description:
+ *      Initializes the thread. Does nothing in all implementations, except
+ *      when statically linked for WIN32.
+ *  Parameters:
+ *      none.
+ *  Returns:
+ *      0 on success, Nonzero on failure.
+ ***************************************************************************/
+static UPNP_INLINE int ithread_initialize_thread(void) {
+	int ret = 0;
+
+#if defined(_WIN32) && defined(PTW32_STATIC_LIB)
+	ret = !pthread_win32_thread_attach_np();
+#endif
+
+	return ret;
+}
+
+
+/****************************************************************************
+ * Function: ithread_cleanup_thread
+ *
+ *  Description:
+ *      Clean up thread resources. Does nothing in all implementations, except
+ *      when statically linked for WIN32.
+ *  Parameters:
+ *      none.
+ *  Returns:
+ *      0 on success, Nonzero on failure.
+ ***************************************************************************/
+static UPNP_INLINE int ithread_cleanup_thread(void) {
+	int ret = 0;
+
+#if defined(_WIN32) && defined(PTW32_STATIC_LIB)
+	ret = !pthread_win32_thread_detach_np();
+#endif
+
+	return ret;
+}
+
 
 /****************************************************************************
  * Function: ithread_mutexattr_init
@@ -170,7 +273,7 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      Initializes a mutex attribute variable.
  *      Used to set the type of the mutex.
  *  Parameters:
- *      ithread_mutexattr_init * attr (must be valid non NULL pointer to 
+ *      ithread_mutexattr_init * attr (must be valid non NULL pointer to
  *                                     pthread_mutexattr_t)
  *  Returns:
  *      0 on success, Nonzero on failure.
@@ -187,7 +290,7 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      Releases any resources held by the mutex attribute.
  *      Currently there are no resources associated with the attribute
  *  Parameters:
- *      ithread_mutexattr_t * attr (must be valid non NULL pointer to 
+ *      ithread_mutexattr_t * attr (must be valid non NULL pointer to
  *                                  pthread_mutexattr_t)
  *  Returns:
  *      0 on success, Nonzero on failure.
@@ -195,19 +298,19 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      See man page for pthread_mutexattr_destroy
  ***************************************************************************/
 #define ithread_mutexattr_destroy pthread_mutexattr_destroy
-  
-  
+
+
 /****************************************************************************
  * Function: ithread_mutexattr_setkind_np
  *
  *  Description:
  *      Sets the mutex type in the attribute.
- *      Valid types are: ITHREAD_MUTEX_FAST_NP 
- *                       ITHREAD_MUTEX_RECURSIVE_NP 
+ *      Valid types are: ITHREAD_MUTEX_FAST_NP
+ *                       ITHREAD_MUTEX_RECURSIVE_NP
  *                       ITHREAD_MUTEX_ERRORCHECK_NP
  *
  *  Parameters:
- *      ithread_mutexattr_t * attr (must be valid non NULL pointer to 
+ *      ithread_mutexattr_t * attr (must be valid non NULL pointer to
  *                                   ithread_mutexattr_t)
  *      int kind (one of ITHREAD_MUTEX_FAST_NP or ITHREAD_MUTEX_RECURSIVE_NP
  *                or ITHREAD_MUTEX_ERRORCHECK_NP)
@@ -216,23 +319,23 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      Returns EINVAL if the kind is not supported.
  *      See man page for pthread_mutexattr_setkind_np
  *****************************************************************************/
-#ifdef PTHREAD_MUTEX_RECURSIVE
+#if defined(PTHREAD_MUTEX_RECURSIVE) || defined(__DragonFly__)
 	#define ithread_mutexattr_setkind_np pthread_mutexattr_settype
 #else
 	#define ithread_mutexattr_setkind_np pthread_mutexattr_setkind_np
-#endif
+#endif /* UPNP_USE_RWLOCK */
 
 /****************************************************************************
  * Function: ithread_mutexattr_getkind_np
  *
  *  Description:
  *      Gets the mutex type in the attribute.
- *      Valid types are: ITHREAD_MUTEX_FAST_NP 
- *                       ITHREAD_MUTEX_RECURSIVE_NP 
+ *      Valid types are: ITHREAD_MUTEX_FAST_NP
+ *                       ITHREAD_MUTEX_RECURSIVE_NP
  *                       ITHREAD_MUTEX_ERRORCHECK_NP
  *
  *  Parameters:
- *      ithread_mutexattr_t * attr (must be valid non NULL pointer to 
+ *      ithread_mutexattr_t * attr (must be valid non NULL pointer to
  *                                   pthread_mutexattr_t)
  *      int *kind (one of ITHREAD_MUTEX_FAST_NP or ITHREAD_MUTEX_RECURSIVE_NP
  *                or ITHREAD_MUTEX_ERRORCHECK_NP)
@@ -241,23 +344,23 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      Always returns 0.
  *      See man page for pthread_mutexattr_getkind_np
  *****************************************************************************/
-#ifdef PTHREAD_MUTEX_RECURSIVE
+#if defined(PTHREAD_MUTEX_RECURSIVE) || defined(__DragonFly__)
 	#define ithread_mutexattr_getkind_np pthread_mutexattr_gettype
 #else
 	#define ithread_mutexattr_getkind_np pthread_mutexattr_getkind_np
-#endif
+#endif /* UPNP_USE_RWLOCK */
+
 
-  
 /****************************************************************************
  * Function: ithread_mutex_init
  *
  *  Description:
  *      Initializes mutex.
  *      Must be called before use.
- *      
+ *
  *  Parameters:
  *      ithread_mutex_t * mutex (must be valid non NULL pointer to pthread_mutex_t)
- *      const ithread_mutexattr_t * mutex_attr 
+ *      const ithread_mutexattr_t * mutex_attr
  *  Returns:
  *      0 on success, Nonzero on failure.
  *      Always returns 0.
@@ -274,14 +377,14 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *  Parameters:
  *      ithread_mutex_t * mutex (must be valid non NULL pointer to pthread_mutex_t)
  *      mutex must be initialized.
- *      
+ *
  *  Returns:
  *      0 on success, Nonzero on failure.
  *      Always returns 0.
  *      See man page for pthread_mutex_lock
  *****************************************************************************/
 #define ithread_mutex_lock pthread_mutex_lock
-  
+
 
 /****************************************************************************
  * Function: ithread_mutex_unlock
@@ -292,7 +395,7 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *  Parameters:
  *      ithread_mutex_t * mutex (must be valid non NULL pointer to pthread_mutex_t)
  *      mutex must be initialized.
- *      
+ *
  *  Returns:
  *      0 on success, Nonzero on failure.
  *      Always returns 0.
@@ -305,9 +408,9 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  * Function: ithread_mutex_destroy
  *
  *  Description:
- *      Releases any resources held by the mutex. 
+ *      Releases any resources held by the mutex.
  *		Mutex can no longer be used after this call.
- *		Mutex is only destroyed when there are no longer any threads waiting on it. 
+ *		Mutex is only destroyed when there are no longer any threads waiting on it.
  *		Mutex cannot be destroyed if it is locked.
  *  Parameters:
  *      ithread_mutex_t * mutex (must be valid non NULL pointer to pthread_mutex_t)
@@ -326,14 +429,16 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *  Description:
  *      Initializes a rwlock attribute variable to default values.
  *  Parameters:
- *      const ithread_rwlockattr_init *attr (must be valid non NULL pointer to 
+ *      const ithread_rwlockattr_init *attr (must be valid non NULL pointer to
  *                                           pthread_rwlockattr_t)
  *  Returns:
  *      0 on success, Nonzero on failure.
  *      Always returns 0.
  *      See man page for pthread_rwlockattr_init
  ***************************************************************************/
-#define ithread_rwlockattr_init pthread_rwlockattr_init
+#if UPNP_USE_RWLOCK
+	#define ithread_rwlockattr_init pthread_rwlockattr_init
+#endif /* UPNP_USE_RWLOCK */
 
 
 /****************************************************************************
@@ -342,26 +447,28 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *  Description:
  *      Releases any resources held by the rwlock attribute.
  *  Parameters:
- *      ithread_rwlockattr_t *attr (must be valid non NULL pointer to 
+ *      ithread_rwlockattr_t *attr (must be valid non NULL pointer to
  *                                  pthread_rwlockattr_t)
  *  Returns:
  *      0 on success, Nonzero on failure.
  *      Always returns 0.
  *      See man page for pthread_rwlockattr_destroy
  ***************************************************************************/
-#define ithread_rwlockattr_destroy pthread_rwlockattr_destroy
-  
-  
+#if UPNP_USE_RWLOCK
+	#define ithread_rwlockattr_destroy pthread_rwlockattr_destroy
+#endif /* UPNP_USE_RWLOCK */
+
+
 /****************************************************************************
  * Function: ithread_rwlockatttr_setpshared
  *
  *  Description:
  *      Sets the rwlock type in the attribute.
- *      Valid types are: ITHREAD_PROCESS_PRIVATE 
+ *      Valid types are: ITHREAD_PROCESS_PRIVATE
  *                       ITHREAD_PROCESS_SHARED
  *
  *  Parameters:
- *      ithread_rwlockattr_t * attr (must be valid non NULL pointer to 
+ *      ithread_rwlockattr_t * attr (must be valid non NULL pointer to
  *                                   ithread_rwlockattr_t)
  *      int kind (one of ITHREAD_PROCESS_PRIVATE or ITHREAD_PROCESS_SHARED)
  *
@@ -370,7 +477,9 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      Returns EINVAL if the kind is not supported.
  *      See man page for pthread_rwlockattr_setkind_np
  *****************************************************************************/
-#define ithread_rwlockatttr_setpshared pthread_rwlockatttr_setpshared
+#if UPNP_USE_RWLOCK
+	#define ithread_rwlockatttr_setpshared pthread_rwlockatttr_setpshared
+#endif /* UPNP_USE_RWLOCK */
 
 
 /****************************************************************************
@@ -378,11 +487,11 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *
  *  Description:
  *      Gets the rwlock type in the attribute.
- *      Valid types are: ITHREAD_PROCESS_PRIVATE 
- *                       ITHREAD_PROCESS_SHARED 
+ *      Valid types are: ITHREAD_PROCESS_PRIVATE
+ *                       ITHREAD_PROCESS_SHARED
  *
  *  Parameters:
- *      ithread_rwlockattr_t * attr (must be valid non NULL pointer to 
+ *      ithread_rwlockattr_t * attr (must be valid non NULL pointer to
  *                                   pthread_rwlockattr_t)
  *      int *kind (one of ITHREAD_PROCESS_PRIVATE or ITHREAD_PROCESS_SHARED)
  *
@@ -391,26 +500,32 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      Always returns 0.
  *      See man page for pthread_rwlockatttr_getpshared
  *****************************************************************************/
-#define ithread_rwlockatttr_getpshared pthread_rwlockatttr_getpshared
+#if UPNP_USE_RWLOCK
+	#define ithread_rwlockatttr_getpshared pthread_rwlockatttr_getpshared
+#endif /* UPNP_USE_RWLOCK */
+
 
-  
 /****************************************************************************
  * Function: ithread_rwlock_init
  *
  *  Description:
  *      Initializes rwlock.
  *      Must be called before use.
- *      
+ *
  *  Parameters:
- *      ithread_rwlock_t * rwlock (must be valid non NULL pointer to pthread_rwlock_t)
- *      const ithread_rwlockattr_t * rwlock_attr 
+ *      ithread_rwlock_t *rwlock (must be valid non NULL pointer to pthread_rwlock_t)
+ *      const ithread_rwlockattr_t *rwlock_attr
  *  Returns:
  *      0 on success, Nonzero on failure.
  *      Always returns 0.
  *      See man page for pthread_rwlock_init
  *****************************************************************************/
-#define ithread_rwlock_init pthread_rwlock_init
-
+#if UPNP_USE_RWLOCK
+	#define ithread_rwlock_init pthread_rwlock_init
+#else
+	/* Read-write locks aren't available: use mutex instead. */
+	#define ithread_rwlock_init ithread_mutex_init
+#endif
 
 /****************************************************************************
  * Function: ithread_rwlock_rdlock
@@ -418,16 +533,20 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *  Description:
  *      Locks rwlock for reading.
  *  Parameters:
- *      ithread_rwlock_t * rwlock (must be valid non NULL pointer to pthread_rwlock_t)
+ *      ithread_rwlock_t *rwlock (must be valid non NULL pointer to pthread_rwlock_t)
  *      rwlock must be initialized.
- *      
+ *
  *  Returns:
  *      0 on success, Nonzero on failure.
  *      Always returns 0.
  *      See man page for pthread_rwlock_rdlock
  *****************************************************************************/
-#define ithread_rwlock_rdlock pthread_rwlock_rdlock
-
+#if UPNP_USE_RWLOCK
+	#define ithread_rwlock_rdlock pthread_rwlock_rdlock
+#else
+	/* Read-write locks aren't available: use mutex instead. */
+	#define ithread_rwlock_rdlock ithread_mutex_lock
+#endif /* UPNP_USE_RWLOCK */
 
 /****************************************************************************
  * Function: ithread_rwlock_wrlock
@@ -435,15 +554,20 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *  Description:
  *      Locks rwlock for writting.
  *  Parameters:
- *      ithread_rwlock_t * rwlock (must be valid non NULL pointer to pthread_rwlock_t)
+ *      ithread_rwlock_t *rwlock (must be valid non NULL pointer to pthread_rwlock_t)
  *      rwlock must be initialized.
- *      
+ *
  *  Returns:
  *      0 on success, Nonzero on failure.
  *      Always returns 0.
  *      See man page for pthread_rwlock_wrlock
  *****************************************************************************/
-#define ithread_rwlock_wrlock pthread_rwlock_wrlock
+#if UPNP_USE_RWLOCK
+	#define ithread_rwlock_wrlock pthread_rwlock_wrlock
+#else
+	/* Read-write locks aren't available: use mutex instead. */
+	#define ithread_rwlock_wrlock ithread_mutex_lock
+#endif /* UPNP_USE_RWLOCK */
 
 
 /****************************************************************************
@@ -453,34 +577,44 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      Unlocks rwlock.
  *
  *  Parameters:
- *      ithread_rwlock_t * rwlock (must be valid non NULL pointer to pthread_rwlock_t)
+ *      ithread_rwlock_t *rwlock (must be valid non NULL pointer to pthread_rwlock_t)
  *      rwlock must be initialized.
- *      
+ *
  *  Returns:
  *      0 on success, Nonzero on failure.
  *      Always returns 0.
  *      See man page for pthread_rwlock_unlock
  *****************************************************************************/
-#define ithread_rwlock_unlock pthread_rwlock_unlock
+#if UPNP_USE_RWLOCK
+	#define ithread_rwlock_unlock pthread_rwlock_unlock
+#else
+	/* Read-write locks aren't available: use mutex instead. */
+	#define ithread_rwlock_unlock ithread_mutex_unlock
+#endif /* UPNP_USE_RWLOCK */
 
 
 /****************************************************************************
  * Function: ithread_rwlock_destroy
  *
  *  Description:
- *      Releases any resources held by the rwlock. 
+ *      Releases any resources held by the rwlock.
  *		rwlock can no longer be used after this call.
- *		rwlock is only destroyed when there are no longer any threads waiting on it. 
+ *		rwlock is only destroyed when there are no longer any threads waiting on it.
  *		rwlock cannot be destroyed if it is locked.
  *  Parameters:
- *      ithread_rwlock_t * rwlock (must be valid non NULL pointer to pthread_rwlock_t)
+ *      ithread_rwlock_t *rwlock (must be valid non NULL pointer to pthread_rwlock_t)
  *      rwlock must be initialized.
  *  Returns:
  *      0 on success. Nonzero on failure.
  *      Always returns 0.
  *      See man page for pthread_rwlock_destroy
  *****************************************************************************/
-#define ithread_rwlock_destroy pthread_rwlock_destroy
+#if UPNP_USE_RWLOCK
+	#define ithread_rwlock_destroy pthread_rwlock_destroy
+#else
+	/* Read-write locks aren't available: use mutex instead. */
+	#define ithread_rwlock_destroy ithread_mutex_destroy
+#endif /* UPNP_USE_RWLOCK */
 
 
 /****************************************************************************
@@ -490,8 +624,8 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      Initializes condition variable.
  *      Must be called before use.
  *  Parameters:
- *      ithread_cond_t * cond (must be valid non NULL pointer to pthread_cond_t)
- *      const ithread_condattr_t * cond_attr (ignored)
+ *      ithread_cond_t *cond (must be valid non NULL pointer to pthread_cond_t)
+ *      const ithread_condattr_t *cond_attr (ignored)
  *  Returns:
  *      0 on success, Nonzero on failure.
  *      See man page for pthread_cond_init
@@ -499,7 +633,6 @@ typedef pthread_rwlock_t ithread_rwlock_t;
 #define ithread_cond_init pthread_cond_init
 
 
-
 /****************************************************************************
  * Function: ithread_cond_signal
  *
@@ -507,7 +640,7 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      Wakes up exactly one thread waiting on condition.
  *      Associated mutex MUST be locked by thread before entering this call.
  *  Parameters:
- *      ithread_cond_t * cond (must be valid non NULL pointer to 
+ *      ithread_cond_t *cond (must be valid non NULL pointer to
  *      ithread_cond_t)
  *      cond must be initialized
  *  Returns:
@@ -524,7 +657,7 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      Wakes up all threads waiting on condition.
  *      Associated mutex MUST be locked by thread before entering this call.
  *  Parameters:
- *      ithread_cond_t * cond (must be valid non NULL pointer to 
+ *      ithread_cond_t *cond (must be valid non NULL pointer to
  *      ithread_cond_t)
  *      cond must be initialized
  *  Returns:
@@ -532,7 +665,7 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      See man page for pthread_cond_broadcast
  *****************************************************************************/
 #define ithread_cond_broadcast pthread_cond_broadcast
-  
+
 
 /****************************************************************************
  * Function: ithread_cond_wait
@@ -542,10 +675,10 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      Associated mutex MUST be locked by thread before entering this call.
  *      Mutex is reacquired when call returns.
  *  Parameters:
- *      ithread_cond_t * cond (must be valid non NULL pointer to 
+ *      ithread_cond_t *cond (must be valid non NULL pointer to
  *      ithread_cond_t)
  *      cond must be initialized
- *      ithread_mutex_t *mutex (must be valid non NULL pointer to 
+ *      ithread_mutex_t *mutex (must be valid non NULL pointer to
  *      ithread_mutex_t)
  *      Mutex must be locked.
  *  Returns:
@@ -553,44 +686,40 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *      See man page for pthread_cond_wait
  *****************************************************************************/
 #define ithread_cond_wait pthread_cond_wait
-  
+
 
   /****************************************************************************
    * Function: pthread_cond_timedwait
    *
-   *  Description:      
-   *      Atomically releases the associated mutex and waits on the condition. 
-   *		If the condition is not signaled in the specified time 
-   *              than the 
-   *		call times out and returns.
-   *		Associated mutex MUST be locked by thread before entering 
-   *              this call.
-   *      Mutex is reacquired when call returns.
+   *	Description:
+   *		Atomically releases the associated mutex and waits on the
+   *	condition.
+   *		If the condition is not signaled in the specified time than the
+   *	call times out and returns.
+   *		Associated mutex MUST be locked by thread before entering this call.
+   *		Mutex is reacquired when call returns.
    *  Parameters:
-   *      ithread_cond_t * cond (must be valid non NULL pointer to 
-   *      ithread_cond_t)
-   *      cond must be initialized
-   *      ithread_mutex_t *mutex (must be valid non NULL pointer to 
-   *      ithread_mutex_t)
-   *      Mutex must be locked.
-   *      const struct timespec *abstime (absolute time, measured 
-   *      from Jan 1, 1970)
+   *      ithread_cond_t *cond (must be valid non NULL pointer to ithread_cond_t)
+   *      	cond must be initialized
+   *      ithread_mutex_t *mutex (must be valid non NULL pointer to ithread_mutex_t)
+   *      	Mutex must be locked.
+   *      const struct timespec *abstime (absolute time, measured from Jan 1, 1970)
    *  Returns:
    *      0 on success. ETIMEDOUT on timeout. Nonzero on failure.
    *      See man page for pthread_cond_timedwait
    ***************************************************************************/
- 
+
 #define ithread_cond_timedwait pthread_cond_timedwait
-  
+
 
   /****************************************************************************
    * Function: ithread_cond_destroy
    *
    *  Description:
-   *      Releases any resources held by the condition variable. 
-   *		Condition variable can no longer be used after this call.	
+   *      Releases any resources held by the condition variable.
+   *		Condition variable can no longer be used after this call.
    *  Parameters:
-   *      ithread_cond_t * cond (must be valid non NULL pointer to 
+   *      ithread_cond_t *cond (must be valid non NULL pointer to
    *      ithread_cond_t)
    *      cond must be initialized.
    *  Returns:
@@ -599,6 +728,65 @@ typedef pthread_rwlock_t ithread_rwlock_t;
    ***************************************************************************/
 #define ithread_cond_destroy pthread_cond_destroy
 
+  /****************************************************************************
+   * Function: ithread_attr_init
+   *
+   *  Description:
+   *      Initialises thread attribute object.
+   *  Parameters:
+   *      ithread_attr_t *attr (must be valid non NULL pointer to
+   *      ithread_attr_t)
+   *  Returns:
+   *      0 on success. Nonzero on failure.
+   *      See man page for pthread_attr_init
+   ***************************************************************************/
+#define ithread_attr_init pthread_attr_init
+
+  /****************************************************************************
+   * Function: ithread_attr_destroy
+   *
+   *  Description:
+   *      Destroys thread attribute object.
+   *  Parameters:
+   *      ithread_attr_t *attr (must be valid non NULL pointer to
+   *      ithread_attr_t)
+   *  Returns:
+   *      0 on success. Nonzero on failure.
+   *      See man page for pthread_attr_destroy
+   ***************************************************************************/
+#define ithread_attr_destroy pthread_attr_destroy
+
+  /****************************************************************************
+   * Function: ithread_attr_setstacksize
+   *
+   *  Description:
+   *      Sets stack size of a thread attribute object.
+   *  Parameters:
+   *      ithread_attr_t *attr (must be valid non NULL pointer to
+   *      ithread_attr_t)
+   *      size_t stacksize (value of stacksize must be greater than
+   *      ITHREAD_STACK_MIN and lower than system-imposed limits
+   *  Returns:
+   *      0 on success. Nonzero on failure.
+   *      See man page for pthread_attr_setstacksize
+   ***************************************************************************/
+#define ithread_attr_setstacksize pthread_attr_setstacksize
+
+  /****************************************************************************
+   * Function: ithread_attr_setdetachstate
+   *
+   *  Description:
+   *      Sets detach state of a thread attribute object.
+   *  Parameters:
+   *      ithread_attr_t *attr (must be valid non NULL pointer to
+   *      ithread_attr_t)
+   *      int detachstate (value of detachstate must be ITHREAD_CREATE_DETACHED
+   *      or ITHREAD_CREATE_JOINABLE)
+   *  Returns:
+   *      0 on success. Nonzero on failure.
+   *      See man page for pthread_attr_setdetachstate
+   ***************************************************************************/
+#define ithread_attr_setdetachstate pthread_attr_setdetachstate
 
   /****************************************************************************
    * Function: ithread_create
@@ -608,7 +796,7 @@ typedef pthread_rwlock_t ithread_rwlock_t;
    *      and argument.
    *  Parameters:
    *      ithread_t * thread (must be valid non NULL pointer to pthread_t)
-   *      ithread_attr_t *attr, IGNORED
+   *      ithread_attr_t *attr
    *      void * (start_routine) (void *arg) (start routine)
    *      void * arg - argument.
    *  Returns:
@@ -632,7 +820,7 @@ typedef pthread_rwlock_t ithread_rwlock_t;
    *      See man page for pthread_cancel
    ***************************************************************************/
 #define ithread_cancel pthread_cancel
-  
+
 
   /****************************************************************************
    * Function: ithread_exit
@@ -646,6 +834,7 @@ typedef pthread_rwlock_t ithread_rwlock_t;
    ***************************************************************************/
 #define ithread_exit pthread_exit
 
+
 /****************************************************************************
    * Function: ithread_get_current_thread_id
    *
@@ -669,44 +858,45 @@ typedef pthread_rwlock_t ithread_rwlock_t;
    ***************************************************************************/
 #define ithread_self pthread_self
 
+
   /****************************************************************************
    * Function: ithread_detach
    *
    *  Description:
-   *		Makes a thread's resources reclaimed immediately 
+   *		Makes a thread's resources reclaimed immediately
    *            after it finishes
-   *            execution.  
+   *            execution.
    *  Returns:
    *		0 on success, Nonzero on failure.
    *      See man page for pthread_detach
    ***************************************************************************/
-#define ithread_detach pthread_detach  
+#define ithread_detach pthread_detach
+
 
   /****************************************************************************
    * Function: ithread_join
    *
    *  Description:
-   *		Suspends the currently running thread until the 
+   *		Suspends the currently running thread until the
    * specified thread
-   *      has finished. 
-   *      Returns the return code of the thread, or ITHREAD_CANCELED 
+   *      has finished.
+   *      Returns the return code of the thread, or ITHREAD_CANCELED
    *      if the thread has been canceled.
    *  Parameters:
    *      ithread_t *thread (valid non null thread identifier)
-   *      void ** return (space for return code) 
+   *      void ** return (space for return code)
    *  Returns:
    *		0 on success, Nonzero on failure.
    *     See man page for pthread_join
    ***************************************************************************/
 #define ithread_join pthread_join
-  
 
 
 /****************************************************************************
  * Function: isleep
  *
  *  Description:
- *		Suspends the currently running thread for the specified number 
+ *		Suspends the currently running thread for the specified number
  *      of seconds
  *      Always returns 0.
  *  Parameters:
@@ -715,17 +905,18 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *		0 on success, Nonzero on failure.
  *              See man page for sleep (man 3 sleep)
  *****************************************************************************/
-#ifndef WIN32
-#define isleep sleep
+#ifdef _WIN32
+	#define isleep(x) Sleep((x)*1000)
 #else
-#define isleep(x) Sleep((x)*1000)
+	#define isleep sleep
 #endif
 
+
 /****************************************************************************
  * Function: isleep
  *
  *  Description:
- *		Suspends the currently running thread for the specified number 
+ *		Suspends the currently running thread for the specified number
  *      of milliseconds
  *      Always returns 0.
  *  Parameters:
@@ -734,36 +925,24 @@ typedef pthread_rwlock_t ithread_rwlock_t;
  *		0 on success, Nonzero on failure.
  *              See man page for sleep (man 3 sleep)
  *****************************************************************************/
-#ifndef WIN32
-#define imillisleep(x) usleep(1000*x)
+#ifdef _WIN32
+	#define imillisleep Sleep
 #else
-#define imillisleep	Sleep
-#endif
-
-#ifdef WIN32
-	#ifndef UPNP_STATIC_LIB
-		#ifdef LIBUPNP_EXPORTS
-			/* set up declspec for dll export to make functions visible to library users */
-			#define EXPORT_SPEC __declspec(dllexport)
-		#else
-			#define EXPORT_SPEC __declspec(dllimport)
-		#endif
-	#else
-		#define EXPORT_SPEC
-	#endif
-#else
-	#define EXPORT_SPEC
+	#define imillisleep(x) usleep(1000*x)
 #endif
 
 
-#ifndef PTHREAD_MUTEX_RECURSIVE
+#if !defined(PTHREAD_MUTEX_RECURSIVE) && !defined(__DragonFly__) && !defined(UPNP_USE_MSVCPP)
+/* !defined(UPNP_USE_MSVCPP) should probably also have pthreads version check - but it's not clear if that is possible */
 /* NK: Added for satisfying the gcc compiler */
 EXPORT_SPEC int pthread_mutexattr_setkind_np(pthread_mutexattr_t *attr, int kind);
 #endif
 
+
 #ifdef __cplusplus
 }
 #endif
 
-#endif /* ITHREADH */
+
+#endif /* ITHREAD_H */
 
